
import pycule
from .quote import Quote
import requests

ENAMINE_CATALOGUES = [
	'BB',
	'SCR',
	'MADE',
	'REAL',
]

import logging
logger = logging.getLogger('HIPPO')

class Quoter:

	def __init__(self,
		supplier: str, 
		username: str = None, 
		password: str = None,
		token: str = None,
	):

		if supplier == 'Enamine':

			self.supplier = 'Enamine'
			self.catalogues = ENAMINE_CATALOGUES
			self.get_quote = self.get_enamine_quote

			assert username
			assert password

			self.wrapper = pycule.core.EnamineWrapper(username=username, password=password)

		elif supplier == 'MCule':

			self.supplier = 'MCule'
			self.get_quote = self.get_mcule_quote
			
			assert token

			self.wrapper = pycule.core.MCuleWrapper(authorisation_token=token)
		
		else:
			logger.error(f'Unsupported supplier: "{supplier}"')
			raise NotImplementedError

	### ENAMINE

	def get_enamine_quote(self, compound):

		try:
		
			smiles = compound.smiles

			logger.header(f'Exact search: {smiles=}')

			for catalogue in self.catalogues:

				logger.header(f'Searching in {self.supplier} {catalogue}...')

				try:
					result = self.wrapper.exactsearch(compound.smiles, catalogue)
				except requests.ConnectionError as e:
					logger.error(f'ConnectionError: {e}')
					return None

				if result['response']['result']['code'] != 0:
					continue

				data = result['response']['data']

				assert len(data) == 1

				data = data[0]

				name = data['Id']

				logger.success(f'Found in {catalogue} w/ id={name}')

				break

			else:

				logger.error('No match found')
				return None
			
			### get the information

			for catalogue in self.catalogues:
				
				logger.header(f'Searching by ID in {self.supplier} {catalogue}...')

				result = self.wrapper.compoundidsearch(name, catalogue=catalogue)

				if result['response']['result']['code'] == 0:
					try:
						logger.success(f'Found in {catalogue}')
						return self.parse_enamine_response(result, compound, catalogue), result
					except NoDataInReponse:
						pass

			else:

				logger.error('No catalog entry')
				return None

		except KeyboardInterrupt:
			logger.warning('Interrupted quoting')

	def pick_enamine_exact_data(self, smiles, data):
	
		if len(data) == 1:
			return data[0]['Id']

		if len(data) == 0:
			raise NoDataInResponse(smiles)

		# sort the results by increasing lead_time, and increasing stereo-complexity
		data = sorted(data, key=lambda x: (self.parse_enamine_delivery_string(x['deliveryDays']), stereo_complexity(x['smile'])))
		mout.warning(f'Picking quickest and least stereo complex result from: {[x["Id"] for x in data]}')
		return data[0]['Id']

	def parse_enamine_response(self, result, compound, catalogue):

		quotes = []

		for i,data in enumerate(result['response']['data']):

			
			if i == 0:

				# update compound metadata
				metadata = dict()
				metadata['alias'] = data['name']
				metadata['cas'] = data['cas']
				metadata['mfcd'] = data['mfcd']
				metadata['formula'] = data['formula']
				metadata['molecular_weight'] = data['mw']
				compound.metadata.update(metadata)

				"""
				do something with: 
				* availability
				* storageCond
				* productUrl
				* lastUpdate
				"""
			
			lead_time = self.parse_enamine_delivery_string(data['deliveryDays'])
			entry = data['Id']
			smiles = data['smile']
			purity = data['purity'] / 100
			
			for pack in data['packs']:
				self.parse_enamine_pack(compound, entry, purity, catalogue, pack, lead_time, smiles)

	def pick_enamine_data(self, comp_id,data):
		
		data = [d for d in data if d['Id'] == comp_id]
		
		if len(data) == 0:
			raise NoDataInReponse

		elif len(data) == 1:
			return data[0]

		data = [d for d in data if not any([p['price'] == 0.0 for p in d['packs']])]

		if len(data) > 1:
			logger.warning(f'Taking first data entry after stripping non-priced ({comp_id})')
		return data[0]

	def parse_enamine_delivery_string(self, string):

		if string.startswith('regular delivery, ') and string.endswith('days'):
			return int(string.removeprefix('regular delivery, ').removesuffix(' days'))

		if string.startswith('backorder, ') and string.endswith('days'):
			return int(string.removeprefix('backorder, ').removesuffix(' days'))

		if string == '3 weeks synthesis time':
			return 15

		raise Exception(f'Unexpected delivery string: {string}')

	def parse_enamine_pack(self,compound, entry, purity, catalogue, pack, lead_time, smiles):

		supplier = 'Enamine'

		match pack['measure']:
			case 'g':
				amount = pack['amount'] * 1000
			case 'mg':
				amount = pack['amount']
			case _:
				raise NotImplementedError(f'Not supported pack measure: {pack["measure"]}')

		price = pack['price']
		currency = pack['currencyName']

		if pack['status'] not in ['Normal', 'Ice pack']:
			print(pack)
			raise Exception(f"{pack['status']=}")

		if not price:
			logger.warning(f'Skipping price-less Enamine pack ({entry})')
			return None

		compound.db.insert_quote(
			compound=compound,
			supplier=supplier,
			catalogue=catalogue,
			entry=entry,
			amount=amount,
			price=price,
			currency=currency,
			purity=purity,
			lead_time=lead_time,
			smiles=smiles,
		)

	### MCULE

	def get_mcule_quote(self, compound):

		try:

			# single exact query

			smiles = compound.smiles

			logger.header(f'Exact search: {smiles=}')

			result = self.wrapper.singlequerysearch(smiles)

			if result['response']['results']:

				results = result['response']['results']

				logger.header(results)

				mcule_ids = [data['mcule_id'] for data in results]
				smiles = [data['smiles'] for data in results]

				logger.success(f'Found w/ ids={mcule_ids}')

				for mcule_id,smile in zip(mcule_ids, smiles):

					# logger.header(f'{self.wrapper.compoundavailability(mcule_id)=}')
					# logger.header(f'{self.wrapper.compounddetails(mcule_id)=}')
					result = self.wrapper.compoundprices(mcule_id)

					for pack in result['response']['best_prices']:
						self.parse_mcule_pack(compound, mcule_id, pack, smile)

					# logger.header(f'{self.wrapper.compoundpricesamount(mcule_id)=}')

			return result

		except KeyboardInterrupt:
			logger.warning('Interrupted quoting')

	def parse_mcule_pack(self, compound, entry, pack, smiles):

		supplier = 'MCule'

		match pack['unit']:
			case 'g':
				amount = pack['amount'] * 1000
			case 'mg':
				amount = pack['amount']
			case _:
				raise NotImplementedError(f'Not supported pack measure: {pack["unit"]}')

		price = pack['price']
		currency = pack['currency']
		purity = pack['purity'] / 100
		lead_time = pack['delivery_time_working_days']

		compound.db.insert_quote(
			compound=compound,
			supplier=supplier,
			catalogue=None,
			entry=entry,
			amount=amount,
			price=price,
			currency=currency,
			purity=purity,
			lead_time=lead_time,
			smiles=smiles,
		)

class NoDataInReponse(Exception):
	...

class NotInCatalogues(Exception):
	...




import pandas as pd

from .cset import CompoundSet
from .pset import PoseSet
from .tags import TagSet
from .rset import ReactionSet
from .compound import Compound
from .reaction import Reaction
from .target import Target
from .pose import Pose

from .db import Database
from pathlib import Path

from tqdm import tqdm

from .tools import inchikey_from_smiles, sanitise_smiles

from mlog import setup_logger
logger = setup_logger('HIPPO')

class HIPPO:
		
	def __init__(self, 
		name: str,  
		db_path: str | Path,
	):

		logger.header('Creating HIPPO animal')

		self._name = name
		# self._target_name = target

		logger.var('name', name, dict(color='arg'))
		# logger.var('target', target, dict(color='arg'))

		if not isinstance(db_path, Path):
			db_path = Path(db_path)
		
		logger.var('db_path', db_path, dict(color='file'))

		self._db_path = db_path
		self._db = Database(self.db_path)

		self._compounds = CompoundSet(self.db, 'compound')
		self._poses = PoseSet(self.db, 'pose')
		self._tags = TagSet(self.db, 'tag')
		self._reactions = ReactionSet(self.db, 'reaction')

		logger.success(f"Initialised animal {self}")
		
	### FACTORIES

	### PROPERTIES

	@property
	def name(self):
		return self._name

	# @property
	# def target_name(self):
	# 	return self._target_name

	@property
	def db_path(self):
		return self._db_path

	@property
	def db(self):
		return self._db

	@property
	def compounds(self):
		return self._compounds

	@property
	def poses(self):
		return self._poses

	@property
	def reactions(self):
		return self._reactions

	@property
	def tags(self):
		return self._tags
	
	### BULK INSERTION

	def add_hits(self, target_name, metadata_csv, aligned_directory, skip=None, debug=False, test=False):

		""" Pass in a generator to the aligned directory from a Fragalysis download"""

		import molparse as mp
		from rdkit.Chem import PandasTools
		from .tools import remove_other_ligands

		# create the target
		target_id = self.db.insert_target(name=target_name)

		if not target_id:
			target_id = self.db.get_target_id(name=target_name)

		skip = skip or []

		logger.var('aligned_directory',aligned_directory)

		count_directories_tried = 0
		count_compound_registered = 0
		count_poses_registered = 0

		meta_df = pd.read_csv(metadata_csv)
		generated_tag_cols = ['ConformerSites', 'CanonSites', 'CrystalformSites', 'Quatassemblies', 'Crystalforms']
		curated_tag_cols = [c for c in meta_df.columns if c not in ['Code', 'Long code', 'Compound code', 'Smiles']+generated_tag_cols]

		for path in tqdm(aligned_directory.iterdir()):

			if not path.is_dir():
				continue

			if path.name in skip:
				continue

			count_directories_tried += 1

			sdfs = list(path.glob('*x?????.sdf'))
			pdbs = list(path.glob('*x?????.pdb'))

			assert len(sdfs) == 1, (path, sdfs)
			assert len(pdbs) == 1, (path, pdbs)

			# logger.var('sdfs[0]',sdfs[0])
			# logger.var('PandasTools', PandasTools.__file__)
			
			# load the SDF
			df = PandasTools.LoadSDF(str(sdfs[0]), molColName='ROMol', idName='ID', strictParsing=True)

			# extract fields
			observation_shortname = path.name.replace('.sdf','')
			observation_longname = df.ID[0]
			mol = df.ROMol[0]
			lig_res_number = int(observation_longname.split('-')[1].split('_')[2])

			# create the single ligand bound pdb

			sys = mp.parse(pdbs[0], verbosity=debug)

			sys = remove_other_ligands(sys, lig_res_number)

			pose_path = str(pdbs[0].resolve()).replace('.pdb','_hippo.pdb')
			smiles = mp.rdkit.mol_to_smiles(mol)

			smiles = sanitise_smiles(smiles, verbosity=debug)

			mp.write(pose_path, sys, shift_name=True, verbosity=debug)

			# create the molecule / pose

			compound_id = self.db.insert_compound(
				# name=crystal_name, 
				smiles=smiles, 
				tags=['hits'],
				warn_duplicate=debug,
			)

			if not compound_id:

				inchikey = inchikey_from_smiles(smiles)
				compound = self.compounds[inchikey]

				if not compound:
					logger.error('Compound exists in database but could not be found by inchikey')
					logger.var('smiles',smiles)
					logger.var('inchikey',inchikey)
					logger.var('observation_shortname',observation_shortname)
					raise Exception
			
			else:
				count_compound_registered += 1
				compound = self.compounds[compound_id]

			# metadata

			meta_row = meta_df[meta_df['Code']==observation_shortname]

			tags = ['hits']
			for tag in curated_tag_cols:
				if meta_row[tag].values[0]:
					tags.append(tag)

			metadata = {}
			for tag in generated_tag_cols:
				metadata[tag] = meta_row[tag].values[0]

			pose_id = self.db.insert_pose(
				compound=compound,
				name=observation_shortname,
				target=target_id,
				path=pose_path,
				tags=tags,
				metadata=metadata,
			)

			if pose_id:
				count_poses_registered += 1

			if test:
				break

		logger.var('#directories parsed', count_directories_tried)
		logger.var('#compounds registered', count_compound_registered)
		logger.var('#poses registered', count_poses_registered)

		return meta_df

	### SINGLE INSERTION

	def register_compound(self, 
		*,
		smiles: str, 
		base: Compound | int | None = None, 
		tags: None | list = None, 
		metadata: None | dict = None,
		return_compound: bool = True,
	) -> Compound:

		"""Use a smiles string to add a compound to the database. If it already exists return the compound"""

		smiles = sanitise_smiles(smiles)
		inchikey = inchikey_from_smiles(smiles)

		compound_id = self.db.insert_compound(
			smiles=smiles, base=base, inchikey=inchikey, tags=tags, 
			metadata=metadata, warn_duplicate=False, commit=True)

		if return_compound or metadata:
			if not compound_id:
				compound = self.compounds[inchikey]
			else:
				compound = self.compounds[compound_id]

			if metadata:
				compound.metadata.update(metadata)

			return compound

		else:
			if not compound_id:
				compound_id = self.db.get_compound_id(name=inchikey)
			return compound_id

	def register_reaction(self,
		*,
		type: str,
		product: Compound,
		reactants: list[Compound | int],
	) -> Reaction:

		### CHECK FOR DUPLICATES
		
		reactant_ids = set(v if isinstance(v, int) else v.id for v in reactants)
		
		for reaction in product.reactions:

			if reaction.type != type:
				continue

			if reaction.product != product:
				continue

			if reaction.reactant_ids != reactant_ids:
				continue

			return reaction

		reaction_id = self.db.insert_reaction(type=type, product=product)

		for reactant in reactants:
			self.db.insert_reactant(compound=reactant, reaction=reaction_id)

		return self.reactions[reaction_id]

	def register_target(self,
		*,
		name: str,
	) -> Target:
		
		target_id = self.db.insert_target(name=name)

		if not target_id:
			target_id = self.db.get_target_id(name=name)

		return self.db.get_target(id=target_id)

	def register_pose(self,
		*,
		compound: Compound | int,
		name: str,
		target: str,
		path: str | None = None,
		reference: int | None = None,
		tags: None | list = None,
		metadata: None | dict = None,
		inspirations: None | list = None,
		return_pose: bool = True,
	) -> Pose:

		longname = f'{compound.name} {target} {name}'
		
		pose_id = self.db.insert_pose(
			compound=compound, longname=longname, name=name, target=target, path=path, 
			tags=tags, metadata=metadata, reference=reference, warn_duplicate=False)

		if return_pose:
			if not pose_id:
				pose = self.poses[longname]
				pose.metadata.update(metadata)
			else:
				pose = self.poses[pose_id]
		else:
			if not pose_id:
				pose_id = self.db.get_pose_id(longname=longname)
			pose = pose_id

		inspirations = inspirations or []
		for inspiration in inspirations:
			self.db.insert_inspiration(original=inspiration, derivative=pose, warn_duplicate=False)

		return pose

	### PLOTTING

	def plot_tag_statistics(self, *args, **kwargs):
		from .plotting import plot_tag_statistics
		return plot_tag_statistics(self, *args, **kwargs)

	def plot_compound_property(self, prop, **kwargs): 
		from .plotting import plot_compound_property
		return plot_compound_property(self, prop, **kwargs)

	def plot_pose_property(self, prop, **kwargs): 
		from .plotting import plot_pose_property
		return plot_pose_property(self, prop, **kwargs)

	def plot_interaction_punchcard(self, poses, subtitle, opacity=1.0, **kwargs):
		from .plotting import plot_interaction_punchcard
		return plot_interaction_punchcard(self, poses=poses, subtitle=subtitle, opacity=opacity, **kwargs)

	### DUNDERS

	def __repr__(self):
		return f'HIPPO("{self.name}")'




import mcol

from rdkit import Chem

from .pose import Pose
from .tags import TagSubset

import logging
logger = logging.getLogger('HIPPO')


class Compound:

	def __init__(self,
			db,
			id: int,
			name: str,
			smiles: str,
			base: int,
			mol: Chem.Mol | bytes | None = None,
			metadata: dict | None = None,
	):
		
		self._id = id
		self._name = name
		self._smiles = smiles
		self._base = base
		self._table = 'compound'
		self._tags = None

		self._metadata = metadata
		
		if isinstance(mol, bytes):
			mol = Chem.Mol(mol)
			
		self._mol = mol

		self._db = db
		
	### FACTORIES

	### PROPERTIES

	@property
	def id(self):
		return self._id
	
	@property
	def name(self):
		return self._name
	
	@property
	def smiles(self):
		return self._smiles
	
	@property
	def mol(self):
		if self._mol is None:
			mol, = self.db.select_where(query='mol_to_binary_mol(compound_mol)', table='compound', key='id', value=self.id, multiple=False)
			self._mol = Chem.Mol(mol)
		return self._mol

	@property
	def metadata(self):
		if self._metadata is None:
			self._metadata = self.db.get_metadata(table='compound', id=self.id)
		return self._metadata

	@property
	def db(self):
		return self._db

	@property
	def tags(self):
		if not self._tags:
			self._tags = self.get_tags()
		return self._tags

	@property
	def poses(self):
		return self.get_poses()

	@property
	def base(self):
		if isinstance(self._base, int):
			self._base = self.db.get_compound(id=self._base)
		return self._base

	@base.setter
	def base(self, b):
		self._base = b
		self.db.update(table='compound', id=self.id, key='compound_base', value=b)

	@property
	def table(self):
		return self._table

	@property
	def reactions(self):
		return self.get_reactions(none=False)
	
	@property
	def dict(self):

		serialisable_fields = ['id','name','smiles']

		data = {}
		for key in serialisable_fields:
			data[key] = getattr(self, key)

		if base := self.base:
			data['base'] = base.name

		if metadata := self.metadata:
			for key in metadata:
				data[key] = metadata[key]

		return data
	
	### METHODS

	def get_tags(self) -> set:
		tags = self.db.select_where(query='tag_name', table='tag', key='compound', value=self.id, multiple=True, none='quiet')
		return TagSubset(self, {t[0] for t in tags})
		# if tags:
		# 	return TagSubset(self, {t[0] for t in tags})
		# else:
		# 	return None

	def get_quotes(self, min_amount=None, supplier=None, max_lead_time=None, none='error', pick_cheapest=False, df=False) -> list[dict]:

		quote_ids = self.db.select_where(query='quote_id', table='quote', key='compound', value=self.id, multiple=True, none=none)

		if quote_ids:
			quotes = [self.db.get_quote(id=q[0]) for q in quote_ids]
		else:
			return []

		if supplier:
			quotes = [q for q in quotes if q.supplier == supplier]

		if min_amount:
			quotes = [q for q in quotes if q.amount >= min_amount]

		if max_lead_time:
			quotes = [q for q in quotes if q.lead_time <= max_lead_time]

		if pick_cheapest:
			return sorted(quotes, key=lambda x: x.price)[0]

		if df:
			from pandas import DataFrame
			return DataFrame([q.asdict() for q in quotes]).drop(columns='compound')
		
		return quotes

	def get_reactions(self, none='error') -> list:

		reaction_ids = self.db.select_where(query='reaction_id', table='reaction', key='product', value=self.id, multiple=True, none=none)

		if reaction_ids:
			reactions = [self.db.get_reaction(id=q[0]) for q in reaction_ids]
		else:
			return []

		if len(reactions) > 1:
			reactions = self.db.prune_reactions(compound=self, reactions=reactions)

		return reactions

	def get_poses(self, 
		target: str = None
	) -> list[Pose]:

		pose_ids = self.db.select_where(query='pose_id', table='pose', key='compound', value=self.id, multiple=True)

		poses = [self.db.get_pose(id=q[0]) for q in pose_ids]

		if target:
			poses = [q for q in poses if q['target'] == target]

		return poses

	def as_ingredient(self, amount, max_lead_time=None, supplier=None):
		
		quote = self.get_quotes(
			pick_cheapest=True, 
			min_amount=amount, 
			max_lead_time=max_lead_time, 
			supplier=supplier,
			none='quiet',
		)
		
		return Ingredient(self, amount, quote, max_lead_time, supplier)

	def draw(self):

		''' draw:

		* compound
		* base
		* inspirations (move to Pose.draw())

		'''

		# print(self.inspirations)
		print(self.base)

		if self.base:
			from molparse.rdkit import draw_mcs
			return draw_mcs({self.base.smiles:str(self), self.smiles:f'{self.base} (base)'})

	### DUNDERS

	def __str__(self):
		return f'C{self.id}'

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{self} "{self.name}"{mcol.unbold}{mcol.ununderline}'

	def __eq__(self, other):
		return self.id == other.id

class Ingredient(Compound):

	"""An ingredient is a Compound with a fixed quanitity and an attached quote"""

	def __init__(self, inherit, amount, quote, max_lead_time=None, supplier=None):
		self._id = inherit.id
		self._name = inherit.name
		self._smiles = inherit.smiles
		self._base = inherit.base			
		self._mol = inherit.mol
		self._db = inherit.db
		
		self._max_lead_time = max_lead_time
		self._supplier = supplier
		self._amount = amount
		self._quote = quote

	@property
	def amount(self):
		return self._amount

	@amount.setter
	def amount(self, a):

		quote = self.get_quotes(
			pick_cheapest=True, 
			min_amount=a, 
			max_lead_time=self._max_lead_time, 
			supplier=self._supplier,
			none='quiet',
		)
		
		self._amount = a

	@property
	def quote(self):
		return self._quote

	def __repr__(self):
		return f'{self.amount:.2f}mg of {mcol.bold}{mcol.underline}{self} "{self.name}"{mcol.unbold}{mcol.ununderline}'

	def __eq__(self, other):

		if self.id != other.id:
			return False

		return self.amount == other.amount




import molparse as mp
import plotly.express as px
import pandas as pd
import mout
import plotly.graph_objects as go

import logging
logger = logging.getLogger('HIPPO')

'''

	ALL GRAPHS DEFINED HERE SHOULD:

	* Have a HIPPO logo
	* Include the target name in the title

'''

import functools

# hippo_graph decorator
def hippo_graph(func):

	@functools.wraps(func)
	def wrapper(animal, *args, **kwargs):

		wrapper_kwargs = {}
		wrapper_keys = ['show', 'html', 'pdf', 'png']
		for key in wrapper_keys:
			wrapper_kwargs[key] = kwargs.pop(key,None)

		fig = func(animal, *args, **kwargs)

		if wrapper_kwargs['show']:
				fig.show()

		if wrapper_kwargs['html']:
			file = wrapper_kwargs['html']
			if not file.endswith('.html'):
				file = f'{file}.html'
			mp.write(file, fig)

		if wrapper_kwargs['pdf']:
			file = wrapper_kwargs['pdf']
			if not file.endswith('.pdf'):
				file = f'{file}.pdf'
			mp.write(file, fig)

		if wrapper_kwargs['png']:
			file = wrapper_kwargs['png']
			if not file.endswith('.png'):
				file = f'{file}.png'
			mp.write(file, fig)

		if not fig.layout.images:
			add_hippo_logo(fig)

		return fig
	
	return wrapper

@hippo_graph
def plot_tag_statistics(animal, color='type', subtitle=None, log_y=False):

	plot_data = []

	for tag in animal.tags.unique:

		num_compounds = len(animal.compounds.get_by_tag(tag=tag))
		data = dict(tag=tag, number=num_compounds, type='compounds')
		plot_data.append(data)

		num_poses = len(animal.poses.get_by_tag(tag=tag))
		data = dict(tag=tag, number=num_poses, type='poses')
		plot_data.append(data)
		
	fig = px.bar(plot_data, x='tag', y='number', color=color, log_y=log_y)

	title = 'Tag Statistics'

	if subtitle:
		title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'
	else:
		title = f'<b>{animal.name}</b>: {title}'
	
	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	fig.update_layout(xaxis_title='Tag', yaxis_title='#')

	return fig

@hippo_graph
def plot_interaction_histogram(animal, poses, feature_metadata, subtitle=None,):

	df = animal._fingerprint_df(poses)

	plot_data = []
	
	for key in df.columns:
		count = int(df[key].sum())

		if not count:
			continue
		
		data = dict(str=key,count=count)

		data['family'] = feature_metadata[key]['family']
		data['res_name'] = feature_metadata[key]['res_name']
		data['res_number'] = feature_metadata[key]['res_number']
		data['res_chain'] = feature_metadata[key]['res_chain']
		data['atom_numbers'] = feature_metadata[key]['atom_numbers']

		data['res_name_number_chain_str'] = f"{feature_metadata[key]['res_name']} {feature_metadata[key]['res_number']} {feature_metadata[key]['res_chain']}"
	
		plot_data.append(data)

	plot_df = pd.DataFrame(plot_data)
	plot_df.sort_values(['res_chain', 'res_number', 'family'], inplace=True)
	plot_df

	fig = px.bar(plot_df, x='res_name_number_chain_str', y='count', color='family', hover_data=plot_df.columns)

	title='Leveraged protein features'

	if subtitle:
		title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'
	else:
		title = f'<b>{animal.name}</b>: {title}'
	
	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Residue')
	fig.update_layout(yaxis_title='#Interactions')

	return fig

@hippo_graph
def plot_interaction_punchcard(animal, poses, subtitle=None, opacity=1.0, group='pose_id'):

	import plotly

	plot_data = []

	categoryarray = {}

	logger.debug(f'{poses=}')

	for pose in poses:

		fingerprint = pose.fingerprint

		if not fingerprint:
			continue

		# loop over each interaction in the pose
		for key, value in fingerprint.items():
			if not value:
				continue
			
			data = dict(str=key,count=value)

			f = animal.db.get_feature(id=key)

			data['id'] = f.id
			data['family'] = f.family
			data['residue_name'] = f.residue_name
			data['residue_number'] = f.residue_number
			data['chain_name'] = f.chain_name
			data['atom_names'] = f.atom_names
			
			data['pose_name'] = pose.name
			data['pose_longname'] = pose.longname
			data['pose_id'] = str(pose)

			data['chain_res_name_number_str'] = f.chain_res_name_number_str

			if data['chain_res_name_number_str'] not in categoryarray:
				categoryarray[data['chain_res_name_number_str']] = (data['chain_res_name_number_str'], f.chain_name, f.residue_number)

			plot_data.append(data)

	plot_df = pd.DataFrame(plot_data)

	plot_df = plot_df.sort_values(group)

	title = 'Interaction Punch-Card'

	if subtitle:
		title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'
	else:
		title = f'<b>{animal.name}</b>: {title}'

	fig = px.scatter(plot_df, x='chain_res_name_number_str', y='family',marginal_x='histogram',marginal_y='histogram', hover_data=plot_df.columns, color=group, title=title)
	
	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Residue', yaxis_title='Feature Family')

	categoryarray = sorted([v for v in categoryarray.values()], key=lambda x: (x[1], x[2]))
	categoryarray = [v[0] for v in categoryarray]

	fig.update_xaxes(categoryorder='array', categoryarray=categoryarray)

	for trace in fig.data:
		if type(trace) == plotly.graph_objs._histogram.Histogram:
			trace.opacity = 1
			trace.xbins.size = 1
		else:
			trace['marker']['size'] = 10
			trace['marker']['opacity'] = opacity

	fig.update_layout(barmode='stack')
	fig.update_layout(scattermode='group', scattergap=0.75)

	# return add_hippo_logo(fig, in_plot=False)
	return add_punchcard_logo(fig)

@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_reactant_amounts(animal, subtitle=None, color='has_price_picker', named_only=False, most_common=None):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	mout.debug('making plot_data')
	plot_data = []
	for bb in bbs:
		d = bb.dict
		# if most_common and d['amount'] is not None:
			
		if not named_only or not d['name_is_smiles']:
			plot_data.append(d)

	if most_common:
		mout.debug('sorting')
		plot_data = sorted(plot_data, key=lambda x: x['amount'], reverse=True)[:most_used_number]

	fig = px.bar(plot_data, x='name', y='amount', color=color, hover_data=plot_data[0].keys())
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Building Blocks'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Reactant', yaxis_title='#Reactions')

	return fig

@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_reactant_price(animal, subtitle=None, amount=20):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	plot_data = []
	for bb in bbs:
		d = bb.dict
		if not d['has_price_picker']:
			continue

		d[f'price_{amount}mg'] = bb.get_price(amount)
		d[f'min_amount'] = bb.price_picker.min_amount

		plot_data.append(d)

	# fig = px.bar(plot_data, x='name', y=f'price_{amount}mg', color='lead_time', log_y=True, hover_data=plot_data[0].keys())
	fig = px.histogram(plot_data, x=f'price_{amount}mg', color='lead_time', hover_data=plot_data[0].keys())
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Reactant Pricing'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(yaxis_title='Number of reactants', xaxis_title=f'Price for {amount}mg [$USD]')

	return fig

@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_reactants_2d(animal, subtitle=None, amount=20):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	plot_data = []
	for bb in bbs:
		d = bb.dict
		if not d['has_price_picker']:
			continue

		d[f'price_{amount}mg'] = bb.get_price(amount)
		d[f'min_amount'] = bb.price_picker.min_amount

		plot_data.append(d)

	fig = px.scatter(plot_data, y='amount', x=f'price_{amount}mg', color='name', hover_data=plot_data[0].keys())
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Building Blocks'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(scattermode='group', scattergap=0.75)

	fig.update_layout(yaxis_title='Quantity [mg]', xaxis_title=f'Price for {amount}mg [$USD]')

	return fig

@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_building_blocks(animal, subtitle=None, color='name_is_smiles'):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	plot_data = []
	for bb in bbs:
		plot_data.append(bb.dict)

	fig = px.scatter(plot_data, x='name', y='max', color='amount')
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Building Blocks'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Reactant', yaxis_title='Quantity')

	return fig

@hippo_graph
def plot_synthetic_routes(animal, subtitle=None, cset='elabs', color='num_reactants'):

	cset = animal.compound_sets[cset]

	plot_data = []
	for reax in cset.reactions:
		plot_data.append(reax.dict)

	# fig = px.bar(plot_data, x='name', y='amount', color=color)
	fig = px.histogram(plot_data, x='type', color=color)

	title = 'Synthetic Routes'

	if not subtitle:
		subtitle = f'"{cset.name}": #compounds={len(cset)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Compound', yaxis_title='#Routes')

	return fig

@hippo_graph
def plot_numbers(animal, subtitle=None):

	'''
		- y-axis: numbers
		- x-categories
			* hits
			* hit poses
			* bases
			* base poses
			* elabs
			* elab poses
			* BBs (total)
			* BBs (in enamine)
	'''

	# cset = animal.compound_sets[cset]

	plot_data = [
		dict(category='Experimental Hits', number=len(animal.hits), type='compound'),
		dict(category='Experimental Hits', number=len(animal.hits.poses), type='poses'),
		dict(category='Base compounds', number=len(animal.bases), type='compound'),
		dict(category='Base compounds', number=len(animal.bases.poses), type='poses'),
		dict(category='Syndirella Elaborations', number=len(animal.elabs), type='compound'),
		dict(category='Syndirella Elaborations', number=len(animal.elabs.poses), type='poses'),
		dict(category='Unique Reactants', number=len(animal.building_blocks), type='compound'),
	]

	fig = px.bar(plot_data, x='category', y='number', log_y=True, color='type')

	title = 'Compounds & Poses'

	title = f'<b>{animal.name}</b>: {title}<br>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	fig.update_layout(xaxis_title=None, yaxis_title='Log(Quantity)')

	return fig

@hippo_graph
def plot_compound_property(animal, prop, style='bar', null=None):

	"""
	Get an arbitrary property from all the compounds in animal.compounds
	
	If one property, plot a 1D histogram
	If 2D plot a bar/scatter

	"""

	if not isinstance(prop, list):
		prop = [prop]

	plot_data = []

	for comp in animal.compounds:

		data = comp.dict

		for p in prop:

			# has attr

			if p not in data:

				# get attr
				if hasattr(comp,p):
					v = getattr(comp,p)
				elif p in comp.metadata:
					v = comp.metadata[p]
				else:
					v = null

				data[p] = v

		plot_data.append(data)

	if len(prop) == 1:

		title = f'Compound {prop[0]}'

		fig = px.histogram(plot_data, x=prop[0])
		
		fig.update_layout(xaxis_title=prop[0], yaxis_title='Quantity')

	elif len(prop) == 2:

		title = f'Compound {prop[0]} vs {prop[1]}'
		
		func = eval(f'px.{style}')
		fig = func(plot_data, x=prop[0], y=prop[1])

		fig.update_layout(xaxis_title=prop[1], yaxis_title=prop[1])

	else:
		mout.error('Unsupported', code='plotting.plot_compound_property.1')

	title = f'<b>{animal.name}</b>: {title}<br>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	return fig

@hippo_graph
def plot_pose_property(animal, prop, poses=None, style='bar', null=None):

	"""
	Get an arbitrary property from all the poses in animal.poses
	
	If one property, plot a 1D histogram
	If 2D plot a bar/scatter

	"""

	if not isinstance(prop, list):
		prop = [prop]

	if not poses:
		poses = animal.poses

	plot_data = []

	for pose in poses:

		data = pose.dict

		for p in prop:

			# has attr

			if p not in data:

				# get attr
				if hasattr(pose,p):
					v = getattr(pose,p)
				elif p in pose.metadata:
					v = pose.metadata[p]
				else:
					v = null

				data[p] = v

		plot_data.append(data)

	if len(prop) == 1:

		title = f'Pose {prop[0]}'

		fig = px.histogram(plot_data, x=prop[0])
		
		fig.update_layout(xaxis_title=prop[0], yaxis_title='Quantity')

	elif len(prop) == 2:

		title = f'Pose {prop[0]} vs {prop[1]}'
		
		func = eval(f'px.{style}')
		fig = func(plot_data, x=prop[0], y=prop[1])

		fig.update_layout(xaxis_title=prop[1], yaxis_title=prop[1])

	else:
		mout.error('Unsupported', code='plotting.plot_pose_property.1')

	title = f'<b>{animal.name}</b>: {title}<br>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	return fig

@hippo_graph
def plot_reactant_sankey(animal, subtitle):

	'''
		BBs (total)
		BBs (in enamine)
		BBs (within budget)
		BBs (within lead-time)
	'''

	total = animal.building_blocks
	quoted = [bb for bb in animal.building_blocks if bb.price_picker is not None]
	lead_time = [bb for bb in quoted if bb.lead_time <= animal.max_lead_time]
	budget = [bb for bb in quoted if bb.get_price(animal.min_bb_quantity) <= animal.max_bb_price]
	bad = [bb for bb in quoted if bb not in lead_time and bb not in budget]
	
	n_total = len(total)
	n_quoted = len(quoted)
	n_quote_not_attempted = len([bb for bb in total if 'quote_attempted' not in bb.tags])
	n_quote_attempted = n_total - n_quote_not_attempted
	n_ok = len([bb for bb in quoted if bb in lead_time and bb in budget])
	# n_bad = len(bad)
	n_expensive = len([bb for bb in quoted if bb not in bad and bb not in budget and bb in lead_time])
	n_slow = len([bb for bb in quoted if bb not in bad and bb in budget and bb not in lead_time])
	n_bad = len(bad)

	print(n_quote_not_attempted)

	labels = [
		f'Total = {n_total}',																# 0
		f'Quote Succeeded = {n_quoted}',														# 1
		f'Quote Failed = {n_quote_attempted - n_quoted}',											# 2
		f'OK = {n_ok}',																			# 3
		f'Too expensive ({animal.min_bb_quantity}mg > ${animal.max_bb_price}) = {n_expensive}',	# 4
		f'Too slow (lead_time > {animal.max_lead_time} days) = {n_slow}',						# 5
		f'Too expensive & slow = {n_bad}',													# 6
		f'Quote not attempted = {n_quote_not_attempted}', # 7
		f'Quote attempted = {n_quote_attempted}' #8
	]

	links = [
		[8, 1, n_quoted], # total --> quote success
		[8, 2, n_quote_attempted - n_quoted], # total --> not quoted
		[1, 3, n_ok], # quoted --> ok
		[1, 4, n_expensive], # quoted --> too expensive
		[1, 5, n_slow], # quoted --> too slow
		[1, 6, len(bad)], # quoted --> bad
		[0, 7, n_quote_not_attempted], # total --> quote_not_attempted
		[0, 8, n_quote_attempted], # total --> quote_attempted
	]

	source = [l[0] for l in links]
	target = [l[1] for l in links]
	value = [l[2] for l in links]

	fig = go.Figure(data=[go.Sankey(
			node = dict(
			# pad = 15,
			# thickness = 20,
			# line = dict(color = "black", width = 0.5),
			label = labels,
			# color = "blue"
		),
			link = dict(
			source = source,
			target = target,
			value = value,
	))])

	fig.update_layout(font_size=10)

	title = 'Reactant Quoting'

	subtitle = animal.reactant_catalog

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	return fig

HIPPO_LOGO_URL = 'https://raw.githubusercontent.com/mwinokan/HIPPO/main/logos/hippo_logo_tightcrop.png'
HIPPO_HEAD_URL = 'https://raw.githubusercontent.com/mwinokan/HIPPO/main/logos/hippo_assets-02.png'

def add_hippo_logo(fig, in_plot=True):

	assert fig.layout.title.text, 'Figure must have a title to add the HIPPO logo'

	if in_plot:

		fig.add_layout_image(dict(
			source=HIPPO_LOGO_URL,
			xref="paper", yref="paper",
			# layer='below',
			
			x=0.95, y=0.95,
			sizex=0.3, sizey=0.3,
			xanchor="right", yanchor="top",
		))

		return fig

	has_legend = fig.layout.legend.title.text is not None
	fig.layout.margin.t = None

	if has_legend:

		fig.add_layout_image(dict(
			source="",
			xref="paper", yref="paper",
			
			x=1, y=1.05,
			sizex=0.4, sizey=0.4,
			xanchor="left", yanchor="bottom",
		))

	else:

		fig.add_layout_image(dict(
			source=HIPPO_LOGO_URL,
			xref="paper", yref="paper",
			
			x=1, y=1.05,
			sizex=0.3, sizey=0.3,
			xanchor="right", yanchor="bottom",
		))
	
	return fig

def add_punchcard_logo(fig):

	fig.add_layout_image(dict(
		source=HIPPO_HEAD_URL,
		xref="paper", yref="paper",
		
		x=1, y=1,
		sizex=0.25, sizey=0.25,
		xanchor="right", yanchor="top",
	))

	return fig



from .db import Database

import mcol

import os

import logging
logger = logging.getLogger('HIPPO')

from .reaction import Reaction

class ReactionSet:

	def __init__(self, 
		db: Database, 
		table: str = 'reaction',
	) -> None:
		
		self._db = db
		self._table = table

	### PROPERTIES

	@property
	def db(self):
		return self._db

	@property
	def table(self):
		return self._table

	### DUNDERS

	def __getitem__(self, key) -> Reaction:
		
		match key:

			case int():

				if key == 0:
					return self.__getitem__(key=1)

				if key < 0:
					key = len(self) + 1 + key
					return self.__getitem__(key=key)

				else:
					return self.db.get_reaction(table=self.table, id=key)

			case str():
				return self.db.get_reaction(table=self.table, name=key)

			case _:
				logger.error(f'Unsupported type for ReactionSet.__getitem__(): {key=} {type(key)}')

		return None

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}set(R x {len(self)}){mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return self.db.count(self.table)

	def __iter__(self):
		return iter(self[i+1] for i in range(len(self)))




import mcol

from .compound import Compound
from .recipe import Recipe

from mlog import setup_logger
logger = setup_logger('HIPPO')

class Reaction:

	def __init__(self,
		db,
		id: int,
		type: str,
		product: int,
		product_yield: float,
	):
		
		self._db = db
		self._id = id
		self._type = type
		self._product = product
		self._product_yield = product_yield
		
	### FACTORIES

	### PROPERTIES

	@property
	def id(self):
		return self._id

	@property
	def type(self):
		return self._type

	@property
	def product(self):
		if isinstance(self._product, int):
			self._product = self.db.get_compound(id=self._product)
		return self._product

	@property
	def product_yield(self):
		return self._product_yield

	@property
	def db(self):
		return self._db

	@property
	def reactants(self):
		from .cset import CompoundSubset
		return CompoundSubset(self.db, indices=self.reactant_ids)

	@property
	def reaction_str(self):
		s = ' + '.join([str(r) for r in self.reactants])
		s = f'{s} -> {str(self.product)}'
		return s

	@property
	def reactant_ids(self):
		return set(v for v in self.get_reactant_ids())

	@property
	def product_id(self):
		return self.product.id
	
	### METHODS

	def get_reactant_amount_pairs(self) -> list[Compound]:

		compound_ids = self.db.select_where(query='reactant_compound, reactant_amount', table='reactant', key='reaction', value=self.id, multiple=True)

		if compound_ids:
			reactants = [(self.db.get_compound(id=id),amount) for id, amount in compound_ids]
		else:
			return []

		return reactants

	def get_reactant_ids(self) -> list[Compound]:

		compound_ids = self.db.select_where(query='reactant_compound', table='reactant', key='reaction', value=self.id, multiple=True)

		if compound_ids:
			return [id for id, in compound_ids]
		else:
			return []

	def get_ingredients(self, amount, return_reactions=False, return_intermediates=False):

		"""recursively assemble a list of ingredients and reactions required to make the compound"""

		ingredients = []
		intermediates = []
		reax = []

		pairs = self.get_reactant_amount_pairs()
		  
		for reactant, reactant_amount in pairs:

			# scale amount
			reactant_amount *= amount #/total_reactant_amount
			reactant_amount /= self.product_yield

			reactions = reactant.get_reactions(none='quiet')

			if reactions:
				assert len(reactions) == 1
				reaction = reactions[0]
				_ingredients, _reactions, _intermediates = reaction.get_ingredients(reactant_amount, return_reactions=True, return_intermediates=True)

				ingredients += _ingredients
				reax += _reactions
				intermediates += _intermediates

				intermediates.append(reaction.product.as_ingredient(reactant_amount))

				reax.append(reaction)

			else:
				ingredient = reactant.as_ingredient(reactant_amount)
				ingredients.append(ingredient)

		output = [ingredients]

		if return_reactions:
			reax.append(self)
			output.append(reax)

		if return_intermediates:
			output.append(intermediates)

		match len(output):
			case 1:
				return output[0]
			case 2:
				return output[0], output[1]
			case 3:
				return output[0], output[1], output[2]

	def get_recipe(self, 
		amount: float = 1, # in mg
	):

		products = [self.product.as_ingredient(amount=amount)]

		reactants, reactions, intermediates = self.get_ingredients(amount=amount, return_reactions=True, return_intermediates=True)

		recipe = Recipe(products=products, reactants=reactants, reactions=reactions, intermediates=intermediates)
		
		return recipe

	def summary(self, amount=1):

		print(f'{self}.id={self.id}')
		print(f'{self}.type={self.type}')
		print(f'{self}.product={self.product}')
		print(f'{self}.product_yield={self.product_yield}')

		reactants = self.get_reactant_amount_pairs()
		print(f'{self}.reactants={reactants}')

		print(f'Ingredients for {amount} mg of product:')
		ingredients = self.get_ingredients(amount=amount)
		print(ingredients)

		return self.get_recipe(amount)

	def draw(self):

		from molparse.rdkit import draw_grid

		reactants = self.reactants

		product = self.product

		mols = [r.mol for r in reactants]
		mols.append(product.mol)

		labels = [f'+ {r}' if i>0 else f'{r}' for i,r in enumerate(reactants)]
		labels.append(f'-> {product}')

		return draw_grid(mols, labels=labels, highlightAtomLists=None)

	### DUNDERS

	def __str__(self):
		return f'R{self.id}'

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{self}: {self.reaction_str} via {self.type}{mcol.unbold}{mcol.ununderline}'

	def __eq__(self, other):

		if self.type != other.type:
			return False

		if self.product != other.product:
			return False

		if self.reactant_ids != other.reactant_ids:
			return False

		return True




import mcol
import molparse as mp

from rdkit import Chem

import numpy as np

from .tags import TagSubset

import pickle

import logging
logger = logging.getLogger('HIPPO')

from molparse.rdkit.features import FEATURE_FAMILIES, COMPLEMENTARY_FEATURES

CUTOFF_PADDING = 1.0

FEATURE_PAIR_CUTOFFS = {
	'Donor Acceptor': 3.5 + CUTOFF_PADDING,
	'Acceptor Donor': 3.5 + CUTOFF_PADDING,
	'NegIonizable PosIonizable': 4.5 + CUTOFF_PADDING,
	'PosIonizable NegIonizable': 4.5 + CUTOFF_PADDING,
	'Aromatic PosIonizable': 4.5 + CUTOFF_PADDING,
	'PosIonizable Aromatic': 4.5 + CUTOFF_PADDING,
	'Aromatic Aromatic': 6.0 + CUTOFF_PADDING,
	'Hydrophobe Hydrophobe': 4.5 + CUTOFF_PADDING,
}

class Pose:

	def __init__(self, 
		db,
		id: int,
		name: str,
		longname: str,
		smiles: str,
		reference: int, # another pose
		path: str,
		compound: int,
		target: str,
		mol: Chem.Mol | bytes | None,
		fingerprint: bytes,
		metadata: dict | None = None,
	):

		self._db = db
		self._id = id
		self._name = name
		self._longname = longname
		self._smiles = smiles
		self._compound_id = compound
		self._target = target
		self._path = path
		self._protein_system = None

		if fingerprint:
			# logger.debug('unpickling fingerprint')
			fingerprint = pickle.loads(fingerprint)		
		self._fingerprint = fingerprint

		# print(f'{self}{metadata=}')
		self._metadata = metadata
		self._tags = None
		self._table = 'pose'
		self._reference = reference

		if isinstance(mol, bytes):
			self._mol = Chem.Mol(mol)
		else: 
			self._mol = mol
		
	### FACTORIES

	### PROPERTIES

	@property
	def db(self):
		return self._db

	@property
	def id(self):
		return self._id

	@property
	def name(self):
		return self._name

	@property
	def longname(self):
		return self._longname

	@property
	def smiles(self):
		return self._smiles

	@property
	def target(self):
		if isinstance(self._target, int):
			self._target = self.db.get_target(id=self._target)
		return self._target

	@property
	def compound_id(self):
		return self._compound_id

	@property
	def compound(self):
		return self.get_compound()

	@property
	def path(self):
		return self._path

	@property
	def reference(self):
		if isinstance(self._reference, int):
			self._reference = self.db.get_pose(id=self._reference)
		return self._reference

	@property
	def mol(self):
		if not self._mol and self.path:

			if self.path.endswith('.pdb'):

				# logger.reading(self.path)
				sys = mp.parse(self.path, verbosity=False)
				
				self.protein_system = sys.protein_system
				
				lig_residues = sys['rLIG']
				if len(lig_residues) > 1:
					logger.warning('Multiple ligands in PDB')
				lig_res = lig_residues[0]
				
				self.mol = lig_res.rdkit_mol

			elif self.path.endswith('.mol'):

				# logger.reading(self.path)
				mol = mp.parse(self.path, verbosity=False)
				self.mol = mol

			else:

				raise NotImplementedError

		return self._mol

	@mol.setter
	def mol(self, m):
		self._mol = m
		self.db.update(table='pose', id=self.id, key='pose_mol', value=m.ToBinary())

	@property
	def protein_system(self):
		if self._protein_system is None and self.path.endswith('.pdb'):
			# logger.debug(f'getting pose protein system {self}')
			self.protein_system = mp.parse(self.path, verbosity=False).protein_system
		return self._protein_system

	@protein_system.setter
	def protein_system(self, a):
		# logger.debug(f'setting {self} protein_system')
		self._protein_system = a

	@property
	def metadata(self):
		if self._metadata is None:
			self._metadata = self.db.get_metadata(table='pose', id=self.id)
		return self._metadata

	@property
	def fingerprint(self):
		return self._fingerprint

	@fingerprint.setter
	def fingerprint(self, fp):

		# remove features that don't exist in this fingerprint?

		self._fingerprint = fp

		# store in the database
		fp = pickle.dumps(fp)
		# logger.debug('pickling fingerprint')
		self.db.update(table=self.table, id=self.id, key=f'{self.table}_fingerprint', value=fp)

	@property
	def tags(self):
		if not self._tags:
			self._tags = self.get_tags()
		return self._tags

	@property
	def inspirations(self):
		return self.get_inspirations()

	@property
	def table(self):
		return self._table

	@property
	def features(self):
		return mp.rdkit.features_from_mol(self.mol)
	
	@property
	def dict(self):

		serialisable_fields = ['id','name','longname','smiles','path','reference']

		data = {}
		for key in serialisable_fields:
			data[key] = getattr(self, key)

		data['compound'] = self.compound.name
		data['target'] = self.target.name

		if metadata := self.metadata:
			for key in metadata:
				data[key] = metadata[key]

		return data

	### METHODS

	def get_compound(self):
		return self.db.get_compound(id=self._compound_id)

	def get_tags(self):
		tags = self.db.select_where(query='tag_name', table='tag', key='pose', value=self.id, multiple=True)
		return TagSubset(self, {t[0] for t in tags})
	
	def get_inspirations(self):
		inspirations = self.db.select_where(query='inspiration_original', table='inspiration', key='derivative', value=self.id, multiple=True, none='quiet')
		
		from .pset import PoseSubset

		if inspirations:
			inspirations = [id for id, in inspirations]
			inspirations = PoseSubset(self.db, indices=inspirations)

		return inspirations

	def calculate_fingerprint(self):

		if self.path.endswith('.pdb'):
			
			import molparse as mp
			protein_system = self.protein_system
			if not self.protein_system:
				# logger.reading(self.path)
				protein_system = mp.parse(self.path, verbosity=False).protein_system

		elif self.path.endswith('.mol') and self.reference:

			logger.debug('fingerprint from .mol and reference pose')
			protein_system = self.reference.protein_system

		else:

			logger.debug('calculate_fingerprint()')
			raise NotImplementedError

		assert protein_system

		comp_features = self.features

		comp_features_by_family = {}
		for family in FEATURE_FAMILIES:
			comp_features_by_family[family] = [f for f in comp_features if f.family == family]

		protein_features = self.target.features
		if not protein_features:
			protein_features = self.target.calculate_features(protein_system)

		fingerprint = {}

		for prot_feature in protein_features:

			prot_family = prot_feature.family

			prot_residue = protein_system.get_chain(prot_feature.chain_name).residues[f'n{prot_feature.residue_number}']

			if not prot_residue:
				continue

			prot_atoms = [prot_residue.get_atom(a).np_pos for a in prot_feature.atom_names.split(' ')]
			
			prot_coord = np.array(np.sum(prot_atoms,axis=0)/len(prot_atoms))

			complementary_family = COMPLEMENTARY_FEATURES[prot_family]

			complementary_comp_features = comp_features_by_family[complementary_family]

			cutoff = FEATURE_PAIR_CUTOFFS[f'{prot_family} {complementary_family}']

			valid_features = [f for f in complementary_comp_features if np.linalg.norm(f - prot_coord) <= cutoff]

			fingerprint[prot_feature.id] = len(valid_features)

		self.fingerprint = fingerprint

	def draw(self):
		
		assert self.inspirations

		from molparse.rdkit import draw_mols

		mols = [self.mol] + [i.mol for i in self.inspirations]

		return draw_mols(mols)

	### DUNDERS

	def __str__(self):
		return f'P{self.id}'

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{self.compound}->{self} "{self.longname}"{mcol.unbold}{mcol.ununderline}'




# from .tools import df_row_to_dict

from .db import Database
from .pose import Pose

import mcol

import os

import logging
logger = logging.getLogger('HIPPO')

class PoseSet:

	def __init__(self,
		db: Database,
		table: str = 'pose',
	) -> None:

		self._db = db
		self._table = table
		
	### FACTORIES

	### PROPERTIES
	
	@property
	def db(self) -> Database:
		return self._db
	
	@property
	def table(self) -> str:
		return self._table

	@property
	def names(self):
		result = self.db.select(table=self.table, query='pose_name', multiple=True)
		return [q for q, in result]

	### METHODS

	def get_by_tag(self,tag):
		values = self.db.select_where(query='tag_pose', table='tag', key='name', value=tag, multiple=True)
		ids = [v for v, in values if v]
		return self[ids]

	### DUNDERS

	def __getitem__(self, key) -> Pose:
		
		match key:

			case int():
				if key == 0:
					return self.__getitem__(key=1)

				if key < 0:
					key = len(self) + 1 + key
					return self.__getitem__(key=key)

				else:
					return self.db.get_pose(table=self.table, id=key)

			case str():
				if len(key.split(' ')) > 1:
					return self.db.get_pose(longname=key)
				else:
					return self.db.get_pose(name=key)

			case list():
				return PoseSubset(self.db, self.table, key)

			case tuple():
				return PoseSubset(self.db, self.table, key)

			case slice():

				start = key.start or 1
				stop = key.stop or len(self)
				step = key.step or 1

				indices = [i for i in range(start, stop, step)]

				return self[indices]

			case _:
				logger.error(f'Unsupported type for PoseSet.__getitem__(): {type(key)}')

		return None

	def __repr__(self) -> str:
		# return f'PoseSet(table="{self.table}")'
		return f'{mcol.bold}{mcol.underline}set(P x {len(self)}){mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return self.db.count(self.table)

	def __iter__(self):
		return iter(self[i+1] for i in range(len(self)))


class PoseSubset(PoseSet):

	def __init__(self,
		db: Database,
		table: str = 'pose',
		indices: list = None,
	):
		self._db = db
		self._table = table

		indices = indices or []

		self._indices = indices

	### PROPERTIES

	@property
	def indices(self):
		return self._indices

	@property
	def names(self):
		return [self.db.select_where(table=self.table, query='pose_name', key='id', value=i, multiple=False)[0] for i in self.indices]

	### DUNDERS

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}subset(P x {len(self)}){mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return len(self.indices)

	def __iter__(self):
		return iter(self.db.get_pose(table=self.table, id=i) for i in self.indices)

	def __getitem__(self, key) -> Pose:
		try:
			index = self.indices[key]
		except IndexError:
			logger.exception(f'list index out of range: {key=} for {self}')
			raise
		return self.db.get_pose(table=self.table, id=index)
